# HCT.pm (HCT::HCT) - the core.
#
# Copyright 2008, 2009
# Slade Maurer
#
# This file is part of HCT, the Hardware Complexity Tool
#
# HCT is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
# 
# HCT is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
# 
# You should have received a copy of the GNU Lesser General Public License
# along with HCT.  If not, see <http://www.gnu.org/licenses/>.

=head1 NAME

HCT::HCT - HCT core.

=head1 DESCRIPTION

=cut

package HCT::HCT;

our $VERSION = '0.7.3';

use strict;
use warnings;

use threads;
use threads::shared;

use POSIX;

use Getopt::Long;
use Time::HiRes;

use HCT::Constants;
use HCT::Std;
use HCT::Std::IO;
use HCT::File;

# The HCT will save old handles of signals for safe.
# Those signals will be replaced and saved. 
our %OLD_SIG = EMPTY_LIST;

# If STDOUT will be redefined, the old handler 
# will be saved
our $OLD_STDOUT = undef;

=head1 SUBROUTINES

=over 4
=cut

=item B<print_usage ()>

Prints help information and exit. Detail information about options can be 
found here: L<http://hct.sourceforge.net/doc/usage.html> 

=cut

sub print_usage
{
	print <<USAGE;
$HCT::NAME
USAGE: [OPTION]... [DIRECTORY/FILE]...

Startup:
  -V,     --version               display the version of HCT and exit.
  -?, -h, --help                  print this help.
          --batch=NAME [DIR/FILE] batch.
  -T,     --use-threads           use threads.
          --max-threads           max threads (if threads defined)
  -q,     --quiet                 quiet (no unnecessary output).
  -a      --list-all              do not ignore entries starting with '.'.
  -R,     --recursive             process subdirectories recursively.

Language options:
          --force-language=LANG   set as default language for input.
          --language=LANG         after this option can be used options
                                  sensetive to language.

Global output options:
  -o,     --output-file=FILE      output to file.
          --output-format=FORMAT  output format: terminal, csv. (by default terminal).

Output options sensetive to language:
          --output-level=NUMBER   output level. (by default 1)          
          --output-exclude=LIST   exclude following list of metrics from report. (separated by commas)
          --output-levels         print output levels. (only if language set)
          --output-metrics        print output metrics. (only if language set)
          --output-no-modules     no modules, only total score.

Debug:
  -d,     --debug                 tool debug.
          --yydebug=NUM           yyparse debug, where num can be:

               01    Token reading (useful for Lexer debugging)
               02    States information
               04    Driver actions (shifts, reduces, accept...)
               08    Parse Stack dump
               16    Error Recovery tracing
               31    Full debugging ouput

Mail bug reports and suggestions to <slade\@computer.org>, <mail\@d2rk.com>.
USAGE

	exit (SUCCESS);
}

=item B<print_version ()>

Prints $HCT::NAME version and exit.

=cut

sub print_version
{
    print <<VERSION;
$HCT::NAME v$HCT::VERSION

You should have received a copy of the GNU Lesser General Public License 
along with HCT. If not, see <http://www.gnu.org/licenses/>.
VERSION
}

=item B<handle_signals ()>

Set up the signal handlers for either parent or child. Each signal 
correspond to special subroutine that should handle it:

	signalname ==> subroutine signalname_signal

All previous handlers of the captured signals will be saved 
into C<%OLD_SIG>.

=cut

sub handle_signals
{
    no strict 'refs';

    for (keys %SIG) {
	my $signame    = $_;
	my $sub_signame = lc $signame; # subroutine signal name
	$sub_signame =~ s {^(?:__)(.*?)(?:__)$} {$1}; 
	my $sub = *{__PACKAGE__ . "::${sub_signame}_signal"};
	# if subroutine for signal is available...
	if (defined &$sub) {
	    # ...save signal handler
	    $OLD_SIG{$signame} = $SIG{$signame};

	    # ...set a new handler
	    $SIG{$signame} = \&$sub;
	}
    }

    use strict 'refs';
}

=item B<die_signal (...)>

Replaces C<__DIE__> signal.

=cut

sub die_signal
{
    HCT::MESSAGE ((caller), "caught die:\n", @_);
}

=item B<restore_signals ()>

Puts captured signals back. Previous signals will be 
loaded from C<%OLD_SIG>.

=cut

sub restore_signals
{
    foreach my $signame (keys %OLD_SIG) {
	$SIG{$signame} = $OLD_SIG{$signame};
    }
}

=item B<handle_stdout ()>

Redirects C<STDOUT>. For instance when user wants to save output data 
in a file it will be redirected to C<HCT::OUTPUT_FILE>. 

The previous C<STDOUT> handler will be saved into C<$OLD_STDOUT>.

=cut

sub handle_stdout
{
    if (defined $HCT::OUTPUT_FILE) {
		HCT::ERROR ("Cannot handle STDOUT: $!\n") unless (open $OLD_STDOUT, '>&', "STDOUT");
		HCT::ERROR ("The output file cannot be opened: $!\n") unless (open STDOUT, '>', $HCT::OUTPUT_FILE);
		HCT::DEBUG ("STDOUT redirected.\n");
    }
	
    elsif ($HCT::OUTPUT_LOCK) {
		HCT::ERROR ("Cannot handle STDOUT: $!\n") unless (open $OLD_STDOUT, '>&', "STDOUT");
		HCT::ERROR ("The output file cannot be opened: $!\n") unless (open STDOUT, '>', '/dev/null');
		HCT::DEBUG ("STDOUT redirected.\n");
    }
}

=item B<restore_stdout ()>

Closes current C<STDOUT> handler and loads previous.

=cut

sub restore_stdout
{
	if (defined $HCT::OUTPUT_FILE) {
		#close STDOUT;

		#HCT::ERROR ("Cannot back STDOUT handler: $!\n")
		#	unless (open STDOUT, ">&", $OLD_STDOUT);
	}
}

=item B<get_options ()>

Gets common and special options.

=cut

sub get_options
{
	# If arguments not defined then print usage and exit
	unless (scalar (@ARGV)) {
		print_usage ();
	}

	# Configure Getopt::Long...
	Getopt::Long::Configure ("pass_through");

	# Read common options
	GetOptions (
	
		# Start up options
	
		"help|h|?"		=> sub { print_usage () },
		"version|V"		=> sub { print_version (); exit (SUCCESS); },
		"quiet"			=> sub { $HCT::VERBOSE = BE_QUIET },
		"use-threads|T"	=> sub { $HCT::USE_THREADS = TRUE },
		"list-all|a"	=> sub { $HCT::LIST_ALL = TRUE },
		"recursive|R"	=> sub { $HCT::LIST_RECURSIVELY = TRUE },

		"max-threads=n"	=> sub {
			HCT::ERROR ("MAX_THREADS can not be less then one!\n") if ($_[1] < 1);
			$HCT::MAX_THREADS = $_[1];
		},
	
		# Global output options
	
		"output-file|o=s"	=> sub { $HCT::OUTPUT_FILE = $_[1] },
		"output-format=s"	=> sub { $HCT::OUTPUT_FORMAT = uc $_[1] },

		# Debug
	
		"debug|d"	=> sub { $HCT::DEBUG = TRUE },
		"yydebug=n"	=> sub { $HCT::YYDEBUG = $_[1] },
		
	);

	# last defined language
	my $lang = undef;

	GetOptions (

		# Language options

		"language=s" => sub {
			$lang = HCT::Lang->uname ($_[1])
		},

		"force-language=s" => sub {
			# Check, does such langauge exists
			HCT::ERROR ("Unknown language: $_[1]\n") unless HCT::Lang->is_defined ($_[1]);
			$HCT::FORCE_LANGUAGE = HCT::Lang->uname ($_[1]);
		},

		# Output options sensetive to language

		"output-metrics" => sub {
			println "$_[0]: language not set." and exit FAILURE unless defined $lang;
			HCT::Lang->get_handler($lang)->output->print_metrics and exit SUCCESS;
		},

		"output-levels" => sub {
			println "$_[0]: language not set." and exit FAILURE unless defined $lang;
			HCT::Lang->get_handler($lang)->output->print_levels and exit SUCCESS;
		},

		"output-exclude=s" => sub {
			my @lang = (defined $lang) ? ($lang) : HCT::Lang->langlist ();
			foreach (@lang) {
				my @metrics = split (/\,/,$_[1]);
				HCT::Lang->get_handler($_)->output->exclude( \@metrics );
			}
		},

		"output-no-modules" => sub {
			my @lang = (defined $lang) ? ($lang) : HCT::Lang->langlist ();
			foreach (@lang) {
				HCT::Lang->get_handler($_)->output->{no_modules} = TRUE;
			}
		},

		"output-level=i" => sub {
			
			if ($_[1] <= 0) {
				HCT::ERROR (__LINE__, "Output level can not be less or equal zero!\n");
			}
			
			my @lang = (defined $lang) ? ($lang) : HCT::Lang->langlist ();
			foreach (@lang) {
				if (!defined HCT::Lang->get_handler($_)->output->{levels}->{$HCT::OUTPUT_FORMAT}->[ $_[1] ]) {
					println "Sorry, language $_ do not support output level $_[1], format $HCT::OUTPUT_FORMAT.";
					exit FAILURE;
				}
				HCT::Lang->get_handler($_)->output->{level} = $_[1];
			}
		},

	);
}

=item B<get_targetlist (TARGETLIST)>

Builds and returns TARGETLIST.

The TARGETLIST is an hash-array reference that consist of TARGETLIST_ITEM's. TARGET_ITEM structure:

	$TARGETLIST_ITEM = { $BATCH_NAME => $BATCH_CONTENT };

	$BATCH_CONTENT = {
		$LANGUAGE => {
			$DIRECTORY => [$FILE]
		}
	};

=cut

sub get_targetlist
{
	my $targetlist = shift;
	my $dirlist;

	# If argument list is empty then print a message
	# and exit
	unless (scalar (@ARGV)) {
		println "Sorry, no task defined, please select a target file or directory or use batch.";
		exit SUCCESS;
	}

	# last defined batch
	my $batch = UNDEFINED_BATCH;

	GetOptions (

		"batch=s" => sub { $batch = $_[1] },

		'<>' => sub {
			my $target = $_[0];
			
			# if target is not a file or directory than skip it...
			unless (is_dir ($target) || is_file ($target)) {
				println "'$target' is not a file or directory." if $HCT::VERBOSE;
				return FALSE;
			}

			$targetlist->{$batch} = { }
				unless defined $targetlist->{$batch};

			my $res = scan_path ($target);
		
			# How many files do we have?
			{
				my $total = 0;
				map { $total += scalar @$_ } values %$res;

				if ($total == 0) {
					println "No files not found for '$target'.";
					return FALSE;
				}
			}

			# Move up by directories and sort files by languages
			DIRECTORY: foreach my $dir (keys %$res) {

				my $filelist = $res->{$dir};

				# if no files in this directory then go to
				# the next directory
				if (scalar (@$filelist) == 0) {
					next DIRECTORY;
				}

				#$targetlist->{$batch}->{$dir} = { }
				#	unless defined $targetlist->{$batch}->{$dir}; 
		
				FILE: foreach my $filename (@$filelist) {

					my $lang = UNKNOWN_LANGUAGE;
					my $ext = get_extension ($filename);

					# if defined force language
					if ( $HCT::FORCE_LANGUAGE ne UNKNOWN_LANGUAGE ) {
						$lang = $HCT::FORCE_LANGUAGE;
					}

					# else try to find correspond language
					elsif (defined (my $associated = $HCT::FILE_ASSOCIATIONS{$ext}) ) {
						$lang = $associated;
					}

					else {
						# FIXME: message or something
					}

					$targetlist->{$batch}->{$lang} = { }
						unless defined $targetlist->{$batch}->{$lang}; 

					unless (defined $targetlist->{$batch}->{$lang}->{$dir}) {
						$targetlist->{$batch}->{$lang}->{$dir} = [$filename]
					}

					# push a new file
					else {
						push @{ $targetlist->{$batch}->{$lang}->{$dir} }, $filename;
					}

				} # foreach

			} # foreach

		},
	);
}

=item B<init ()>

Initializes the tool: captures signals for handling, initializes HDL's, 
gets common and special options, captures C<STDOUT> if needed.

=cut

sub init
{
	# Capture signals for handling.
	handle_signals ();

	# Initialize HDLs.
	HCT::Lang->init ();

	# Get options HCT options: common and special.
	get_options ();

	#if (is_interactive ()) {
#		$HCT::FRONTEND->run ();
	#}

	# Handle STDOUT if needed.
	handle_stdout ();
}

sub process_file
{
	# if we have threads support
	if ($HCT::USE_THREADS) {
		$SIG{'KILL'} = sub {
			threads->exit ();
		};
	}

	my ($dir, $lang, $filename) = (shift, shift, shift);

	# Get a new analyzer for file
	my $clone = HCT::Lang->get_handler ($lang)->clone ();

	# Wow!
	unless ( defined $clone ) {
		return EMPTY_LIST;
	}

	local *FH;

	# if file exists
	if (is_file ($filename)) {

		# if we cannot open it
		unless (open FH, $filename) {
			HCT::DEBUG ("File '$filename' cannot be opened: $!\n");
			return EMPTY_LIST;
		}

	}

	$clone->input (\*FH);
	$clone->parse;
	$clone->compute_metrics;

	# XXX:	we wont to return the whole object, just the
	#		result that we need.

	return (
		$dir,
		$lang,
		$filename,
		$clone->{metrics}
	);
}

sub run_multiengine ($$)
{
	my ($targetlist, $resultlist) = (shift, shift);

	# push temporary result to result list
	my $save_result = sub
	{
		# Wow! No result!
		if (scalar (@_) == 0) {
			return;
		}

		my ($dir, $lang, $filename, $result) = @_;
		$resultlist->{$dir}->{$lang}->{$filename} = $result;
	};

	my $total = 0; # total number of passed files
	my %threads = (); # list of threads 

	# Look up directory, take language and corresponding filelist.
	# Skip if language unknown.

	while (my ($dir,$langlist) = each %$targetlist) {
			
		while (my ($lang, $dirlist) = each %$langlist) {
			next
				if $lang eq UNKNOWN_LANGUAGE;

			$resultlist->{$dir}->{$lang} = { };

			my $filelist = [];
			map { push @$filelist, @$_ } values %$dirlist;

			my $numfiles = scalar (@$filelist);
			my $index = 0;

			# while list not empty
			while ($index < $numfiles) {

				# Each thread consist of two parts: thread handler and start time 

				# If we can run a new thread
				if ( $HCT::MAX_THREADS > 0
					&& scalar(keys %threads) <= $HCT::MAX_THREADS ) {
						my $filename = $filelist->[$index];
	
						$threads{$total} = [
							threads->new (
								{
									context => 'list'
								},
								\&process_file,		# +-> subroutine (directory, language, filename)
								$dir,				# +-------------------^         ^          ^
								$lang,				# +-----------------------------+          |
								$filename,			# +----------------------------------------+
							),
							time
						];

						$total += 1;
						$index += 1;							
				}

				# Else check up what thread is finished and what is timeouted.
				else {
						
					THREAD:
					foreach (keys %threads) {
							
						# find out what thread timeouted
						if ( $HCT::THREAD_TIMEOUT != 0 ) {
							if ((time - $threads{$_}->[1]) > $HCT::THREAD_TIMEOUT) {
								$threads{$_}->[0]->kill('KILL')->detach ();
								delete $threads{$_};
								next THREAD;
							}
						}
							
						# find out what thread finished
						if ( $threads{$_}->[0]->is_joinable() ) {
							$save_result->($threads{$_}->[0]->join ());
							delete $threads{$_};
						}
							
					}
						
				} # else

			} # while
				
		} # while

	} # while

	# now wait until all threads will be finished
	foreach (keys %threads) {
		$save_result->($threads{$_}->[0]->join());
		delete $threads{$_};
	}

}

sub run_stdengine ($$)
{
	my ($targetlist, $resultlist) = (shift, shift);

	# push temporary result to result list
	my $save_result = sub {
		my ($dir, $lang, $filename, $result) = @_;
		$resultlist->{$dir}->{$lang}->{$filename} = $result;
	};

	while (my ($batch, $langlist) = each %$targetlist) {

		while (my ($lang, $dirlist) = each %$langlist) {

			while (my ($dir, $filelist) = each %$dirlist) {

				next if $lang eq UNKNOWN_LANGUAGE;
		
				# new filelist
				$resultlist->{$dir}->{$lang} = { };

				foreach my $filename (@$filelist) {
					$save_result->(&process_file($dir, $lang, $filename));
				}

			} # while

		} # while

	} # while
}

sub do_main_loop
{
	my %targetlist = EMPTY_LIST;
	my %resultlist = EMPTY_LIST;

	# Get target list
	get_targetlist (\%targetlist);

	# If we have to use threads then run multi-engine mechanism
	if ($HCT::USE_THREADS == TRUE) {
		run_multiengine (\%targetlist,\%resultlist);
	}

	# else run standard engine
	else {
		run_stdengine (\%targetlist,\%resultlist);
	}

	do_output (\%resultlist);
}

sub do_output
{
	my ($resultlist) = (shift);
	
	while ( my ($dir, $langlist) = each %$resultlist ) {
		
		if ( $HCT::BATCH == TRUE ) {
			print "Batch: "
			. ( scalar keys %$langlist )
			. " language(s), "
			. ( scalar map { keys %$_ } values %$langlist )
			. " file(s)\n\n";
		}
		else {
			printf "Directory: $dir\n\n";
		}
		
		while ( my ($lang, $filelist) = each %$langlist ) {
			printf("%s, %d file(s)\n",$lang,scalar keys %$filelist);
			HCT::Lang->get_handler($lang)->output->output ($lang, $filelist);
		}
		
		println;
	}
	
}

sub main
{
	&init;
	&do_main_loop;
	&cleanup;
}

sub cleanup
{
	# Restore signals
	&restore_signals;
	
	# Restore STDOUT
	&restore_stdout;
}

1;

=back
=cut

